package soma;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

/**
 * Solveur de puzzle de SOMA
 * 
 * @author xphilippeau
 */
public class Solver {

	private class Candidate {
		SomaElement soma;
		List<Shape> shapes = new ArrayList<Shape>();
		public Candidate(SomaElement soma) {
			this.soma = soma;
		}
	}
	
	@SuppressWarnings("serial")
	private class Candidates extends LinkedList<Candidate> {}

	@SuppressWarnings("serial")
	private class Solution extends HashMap<SomaElement,Shape> {}
	
	// on initialise la liste des formes qui composeront la solution
	private Solution solution = new Solution();
	
	/**
	 * @param puzzledata int[N][3] liste des coordonnes des N cubes composant le puzzle
	 * @return true si le puzzle est rsoluble
	 */
	public boolean solve(Puzzle puzzle) {

		// on initialise la liste des candidats  la rsolution du puzzle
		Candidates candidates = new Candidates();

		// pour chaque element SOMA
		for(SomaElement element : SomaElement.values()) {
			Candidate candidate = new Candidate(element);
			
			// calcule toutes les positions/orientations possibles de l'element
			for(RotationMatrix r : RotationMatrix.values())
				for(int x=puzzle.minX;x<=puzzle.maxX;x++)
					for(int y=puzzle.minY;y<=puzzle.maxY;y++)
						for(int z=puzzle.minZ;z<=puzzle.maxZ;z++) {
							
							Shape s = Transform.transform(element.data, x, y, z, r);
							if (s==null) continue;

							// si l'element ainsi positionn est inclu dans le puzzle
							// on l'ajoute  la liste des positions possibles
							if (!puzzle.contains(s)) continue;
							if (Shape.contains(candidate.shapes, s)) continue;
							candidate.shapes.add(s);
						}
			
			//System.out.println(element.name()+" possibilties="+candidate.shapes.size());
			
			// On ajoute le candidat  la liste
			candidates.add(candidate);
			
		}
		
		// on trie les candidats pour commencer par ceux avec le minimum de positions possibles
		Collections.sort(candidates, new Comparator<Candidate>() {
			public int compare(Candidate c1, Candidate c2) {
				return (c1.shapes.size() - c2.shapes.size());
			}
		});
		
		// on explore recursivement l'arbre des solutions
		solution.clear();
		boolean result = explore(candidates,puzzle);
		
		// on retourne le rsultat de l'exploration
		return result;
	}
	
	// exploration de l'arbre des solutions
	private boolean explore(Candidates candidates, Puzzle puzzle) {
	
		// Plus de candidats : on regarde l'tat du puzzle 
		if (candidates.isEmpty()) {
			// retourne "true" si le puzzle est rsolu
			return puzzle.isEmpty();
		}
		
		// extrait le prochain candidat
		Candidate candidate = candidates.removeFirst();
		
		// teste chaque forme (pr-positione) de ce candidat
		for(Shape s : candidate.shapes) {
			// si la forme n'est pas contenu dans le puzzle actuel -> suivant
			if (!puzzle.contains(s)) continue;
			
			// ajoute la forme a la solution et met  jour le puzzle
			solution.put(candidate.soma,s);
			puzzle.remove(s);
			// exploration recursive. On sort (SUCESS) si le reste du puzzle est resoluble
			if (explore(candidates,puzzle)) return true;
			// sinon... backtrack.
			solution.remove(candidate.soma);
			puzzle.add(s);
		}
		
		// Impossible de rsoudre le puzzle actuel. backtrack + Sortie (FAIL)
		candidates.addFirst(candidate);
		return false;
	}
	
	/**
	 * @return la solution trouve sous forme textuelle
	 */
	public String getSolutionTxt() {
		StringBuilder sb = new StringBuilder();
		for(Entry<SomaElement,Shape> entry:solution.entrySet()) {
			sb.append(entry.getKey().name()).append(" : ");
			for(int cube:entry.getValue().cubes) 
				sb.append(Arrays.toString(Coordinates.ToCoord(cube))).append(" ");
			sb.append("\n");
		}
		return sb.toString();
	}
	
	public Map<SomaElement,Shape> getSolution() {
		return this.solution;
	}
}
